from libc.stddef cimport wchar_t

cdef extern from "MinHook.h":
    ctypedef void* LPVOID
    ctypedef const char* LPCSTR;
    ctypedef const wchar_t* LPCWSTR;

    ctypedef enum MH_STATUS:
        # Unknown error. Should not be returned.
        MH_UNKNOWN

        # Successful.
        MH_OK

        # MinHook is already initialized.
        MH_ERROR_ALREADY_INITIALIZED

        # MinHook is not initialized yet, or already uninitialized.
        MH_ERROR_NOT_INITIALIZED

        # The hook for the specified target function is already created.
        MH_ERROR_ALREADY_CREATED

        # The hook for the specified target function is not created yet.
        MH_ERROR_NOT_CREATED

        # The hook for the specified target function is already enabled.
        MH_ERROR_ENABLED

        # The hook for the specified target function is not enabled yet, or already
        # disabled.
        MH_ERROR_DISABLED

        # The specified pointer is invalid. It points the address of non-allocated
        # and/or non-executable region.
        MH_ERROR_NOT_EXECUTABLE

        # The specified target function cannot be hooked.
        MH_ERROR_UNSUPPORTED_FUNCTION

        # Failed to allocate memory.
        MH_ERROR_MEMORY_ALLOC

        # Failed to change the memory protection.
        MH_ERROR_MEMORY_PROTECT

        # The specified module is not loaded.
        MH_ERROR_MODULE_NOT_FOUND

        # The specified function is not found.
        MH_ERROR_FUNCTION_NOT_FOUND

    enum:
        MH_ALL_HOOKS

    MH_STATUS MH_Initialize()

    MH_STATUS MH_Uninitialize()

    MH_STATUS MH_CreateHook(LPVOID pTarget, LPVOID pDetour, LPVOID *ppOriginal)

    MH_STATUS MH_CreateHookApi(
        LPCWSTR pszModule, LPCSTR pszProcName, LPVOID pDetour, LPVOID *ppOriginal)

    MH_STATUS MH_CreateHookApiEx(
        LPCWSTR pszModule, LPCSTR pszProcName, LPVOID pDetour, LPVOID *ppOriginal, LPVOID *ppTarget)

    MH_STATUS MH_RemoveHook(LPVOID pTarget)

    MH_STATUS MH_EnableHook(LPVOID pTarget)

    MH_STATUS MH_DisableHook(LPVOID pTarget)

    MH_STATUS MH_QueueEnableHook(LPVOID pTarget)

    MH_STATUS MH_QueueDisableHook(LPVOID pTarget)

    MH_STATUS MH_ApplyQueued()

    const char * MH_StatusToString(MH_STATUS status)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'cyminhook'
copyright = '2022, Segev Finer'
author = 'Segev Finer'

# The full version, including alpha/beta/rc tags
import cyminhook
release = cyminhook.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.githubpages'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

html_theme_options = {
    'github_button': True,
    'github_user': 'segevfiner',
    'github_repo': 'cyminhook',
    'github_type': 'star',
    'extra_nav_links': {
        'cypcap @ PyPI': 'https://pypi.org/project/cyminhook/',
        'cypcap @ GitHub': 'https://github.com/segevfiner/cyminhook',
    }
}


# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}


def remove_bpf_data_docstring(app, what, name, obj, options, lines):
    if what == 'data' and name.startswith("cypcap.bpf"):
        lines.clear()


def setup(app):
    app.connect("autodoc-process-docstring", remove_bpf_data_docstring)



import ctypes
import ctypes.wintypes
import cyminhook
import win32api
import win32con


class MessageBoxExWHook(cyminhook.MinHook):
    signature = ctypes.WINFUNCTYPE(
        ctypes.c_int,
        ctypes.wintypes.HWND,
        ctypes.wintypes.LPCWSTR,
        ctypes.wintypes.LPCWSTR,
        ctypes.wintypes.UINT,
        ctypes.wintypes.WORD,
        use_last_error=True,
    )

    target = ctypes.windll.user32.MessageBoxExW

    def detour(self, hWnd, lpText, lpCaption, uType, langId):
        return self.original(hWnd, "Hooked", "Hooked", uType, langId)


with MessageBoxExWHook() as hook:
    hook.enable()

    win32api.MessageBox(None, "Hello, World!", "Python", win32con.MB_OK)



import ctypes
import ctypes.wintypes
import cyminhook
import win32api
import win32con

signature = ctypes.WINFUNCTYPE(
    ctypes.c_int,
    ctypes.wintypes.HWND,
    ctypes.wintypes.LPCWSTR,
    ctypes.wintypes.LPCWSTR,
    ctypes.wintypes.UINT,
    ctypes.wintypes.WORD,
    use_last_error=True,
)

target = ctypes.windll.user32.MessageBoxExW

def detour(hWnd, lpText, lpCaption, uType, langId):
    return hook.original(hWnd, "Hooked", "Hooked", uType, langId)

with cyminhook.MinHook(signature=signature, target=target, detour=detour) as hook:
    hook.enable()

    win32api.MessageBox(None, "Hello, World!", "Python", win32con.MB_OK)



# cython: language_level=3str, binding=True
"""
Hook functions on Windows using MinHook.
"""
import enum
import ctypes

from libc.stdint cimport uintptr_t
from . cimport cminhook


class Status(enum.IntEnum):
    """MinHook status codes."""
    MH_UNKNOWN = cminhook.MH_UNKNOWN
    MH_OK = cminhook.MH_OK
    MH_ERROR_ALREADY_INITIALIZED = cminhook.MH_ERROR_ALREADY_INITIALIZED
    MH_ERROR_NOT_INITIALIZED = cminhook.MH_ERROR_NOT_INITIALIZED
    MH_ERROR_ALREADY_CREATED = cminhook.MH_ERROR_ALREADY_CREATED
    MH_ERROR_NOT_CREATED = cminhook.MH_ERROR_NOT_CREATED
    MH_ERROR_ENABLED = cminhook.MH_ERROR_ENABLED
    MH_ERROR_DISABLED = cminhook.MH_ERROR_DISABLED
    MH_ERROR_NOT_EXECUTABLE = cminhook.MH_ERROR_NOT_EXECUTABLE
    MH_ERROR_UNSUPPORTED_FUNCTION = cminhook.MH_ERROR_UNSUPPORTED_FUNCTION
    MH_ERROR_MEMORY_ALLOC = cminhook.MH_ERROR_MEMORY_ALLOC
    MH_ERROR_MEMORY_PROTECT = cminhook.MH_ERROR_MEMORY_PROTECT
    MH_ERROR_MODULE_NOT_FOUND = cminhook.MH_ERROR_MODULE_NOT_FOUND
    MH_ERROR_FUNCTION_NOT_FOUND = cminhook.MH_ERROR_FUNCTION_NOT_FOUND


class Error(Exception):
    """MinHook error."""
    def __init__(self, status):
        super().__init__(status)
        self.status = Status(status)


status = cminhook.MH_Initialize()
if status != cminhook.MH_OK:
    raise Error(status)


cdef class MinHook:
    """
    MinHook hook.

    *signature* is a :mod:`ctypes` function signature created by either :func:`ctypes.CFUNCTYPE`,
    :func:`ctypes.WINFUNCTYPE`, :func:`ctypes.PYFUNCTYPE`. *target* is either the address to hook
    or a :mod:`ctypes` function object for the function to hook. *detour* is the Python callable
    that will be called by the hook.

    .. warning:: Be careful not to enter an infinite recursion from the *detour* function.
    """
    cdef public object signature
    cdef public object target
    cdef cminhook.LPVOID _target
    cdef public object detour
    cdef object _detour
    cdef cminhook.LPVOID _original
    cdef readonly object original

    def __init__(self, *, signature=None, target=None, detour=None):
        if signature is None:
            self.signature = getattr(self, 'signature', None)
            if self.signature is None:
                raise ValueError("signature not specified")
        else:
            self.signature = signature

        if target is None:
            self.target = getattr(self, 'target', None)
            if self.target is None:
                raise ValueError("target not specified")
        else:
            self.target = target

        if hasattr(self.target, 'argtypes'):
            self.target = ctypes.cast(self.target, ctypes.c_void_p).value

        self._target = <cminhook.LPVOID><uintptr_t>self.target

        if detour is None:
            self.detour = getattr(self, 'detour', None)
            if self.detour is None:
                raise ValueError("detour not specified")
        else:
            self.detour = detour

        self._detour = self.signature(self.detour)

        status = cminhook.MH_CreateHook(
            self._target,
            <cminhook.LPVOID><uintptr_t>ctypes.cast(self._detour, ctypes.c_void_p).value,
            &self._original
        )
        if status != cminhook.MH_OK:

            raise Error(status)

        self.original = self.signature(<uintptr_t>self._original)

    def __dealloc__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    cdef close(self):
        """Close the hook. Removing it."""
        if self._original is not NULL:
            cminhook.MH_RemoveHook(self._target)
            self._original = NULL

    cdef _check_closed(self):
        if self._original is NULL:
            raise ValueError("Operation on closed Pcap")

    def enable(self):
        """Enable the hook."""
        self._check_closed()

        status = cminhook.MH_EnableHook(self._target)
        if status != cminhook.MH_OK:
            raise Error(status)

    def disable(self):
        """Disable the hook."""
        self._check_closed()

        status = cminhook.MH_DisableHook(self._target)
        if status != cminhook.MH_OK:
            raise Error(status)


def queue_enable(MinHook hook not None):
    """Queue to enable an already created hook."""
    status = cminhook.MH_QueueEnableHook(hook._target)
    if status != cminhook.MH_OK:
        raise Error(status)


def queue_disable(MinHook hook not None):
    """Queue to disable an already created hook."""
    status = cminhook.MH_QueueDisableHook(hook._target)
    if status != cminhook.MH_OK:
        raise Error(status)


def apply_queued():
    """Apply all queued changes in one go."""
    status = cminhook.MH_ApplyQueued()
    if status != cminhook.MH_OK:
        raise Error(status)



import re
from skbuild import setup
from setuptools import find_packages


with open("cyminhook/__init__.py", "r", encoding="utf-8") as f:
    version = re.search(r'(?m)^__version__ = "([a-zA-Z0-9.-]+)"', f.read()).group(1)

with open("README.rst", "r", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="cyminhook",
    version=version,
    author="Segev Finer",
    author_email="segev208@gmail.com",
    description="Hook functions on Windows using MinHook",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/segevfiner/cyminhook",
    project_urls={
        "Documentation": "https://segevfiner.github.io/cyminhook/",
        "Issue Tracker": "https://github.com/segevfiner/cyminhook/issues",
    },
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Win32 (MS Windows)",
        "Intended Audience :: Developers",
        "Topic :: Security",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="MinHook",
    zip_safe=False,
    packages=find_packages(),
    python_requires='>=3.6',
    extras_require={
        "dev": [
            "sphinx==4.*",
        ],
    },
)
